/*
 * Copyright (c) 2007 MIPS Technologies, Inc.
 * All rights reserved.
 * 
 * Unpublished rights (if any) reserved under the copyright laws of the
 * United States of America and other countries.
 * 
 * This code is proprietary to MIPS Technologies, Inc. ("MIPS
 * Technologies").  Any copying, reproducing, modifying or use of
 * this code (in whole or in part) that is not expressly permitted
 * in writing by MIPS Technologies or an authorized third party is
 * strictly prohibited.  At a minimum, this code is protected under
 * unfair competition and copyright laws.  Violations thereof may result
 * in criminal penalties and fines.
 * 
 * MIPS Technologies reserves the right to change this code to improve
 * function, design or otherwise.  MIPS Technologies does not assume any
 * liability arising out of the application or use of this code, or of
 * any error or omission in such code.  Any warranties, whether express,
 * statutory, implied or otherwise, including but not limited to the
 * implied warranties of merchantability or fitness for a particular
 * purpose, are excluded.  Except as expressly provided in any written
 * license agreement from MIPS Technologies or an authorized third party,
 * the furnishing of this code does not give recipient any license to
 * any intellectual property rights, including any patent rights, that
 * cover this code.
 * 
 * This code shall not be exported, reexported, transferred, or released,
 * directly or indirectly, in violation of the law of any country or
 * international law, regulation, treaty, Executive Order, statute,
 * amendments or supplements thereto.  Should a conflict arise regarding
 * the export, reexport, transfer, or release of this code, the laws of
 * the United States of America shall be the governing law.
 * 
 * This code constitutes one or more of the following: commercial
 * computer software, commercial computer software documentation or
 * other commercial items.  If the user of this code, or any related
 * documentation of any kind, including related technical data or
 * manuals, is an agency, department, or other entity of the United
 * States government ("Government"), the use, duplication, reproduction,
 * release, modification, disclosure, or transfer of this code, or
 * any related documentation of any kind, is restricted in accordance
 * with Federal Acquisition Regulation 12.212 for civilian agencies
 * and Defense Federal Acquisition Regulation Supplement 227.7202 for
 * military agencies.  The use of this code by the Government is further
 * restricted in accordance with the terms of the license agreement(s)
 * and/or applicable contract terms and conditions covering this code
 * from MIPS Technologies or an authorized third party.
 */

// Includes:
// MT ASE support where you will find the mips_mt* and mips32_mt* macros
#include <mips/mt.h>
#define mips32_mt_gettcschedule()          _m32c0_mftc0(2,6)
#define mips32_mt_settcschedule(val)         _m32c0_mttc0(2,6,val)

// ITC defines not currently in include files
#define ERRCTL_ITC            0x04000000  /* ITC select for cache opts in ErrCtl register (26,0) */
#define ITC_BypassView        0x00000000
#define ITC_ControlView       0x00000008
#define ITC_EmptyFullSyncView 0x00000010
#define ITC_EmptyFullTryView  0x00000018
#define ITC_PVSyncView        0x00000020
#define ITC_PVTryView         0x00000028
#define ITC_En                0x00000001
#define ITC_E                 0x00000001

// ITC configuration defines
// ITC Block must be addressed aligned on a 512 boundry to allow for View bits
unsigned int ITC_Block[1024] __attribute__((aligned(1024)));
unsigned int *ITC_BlockNC;
unsigned int *ITC_FirstPVCell;
#define ITC_NumEntries 8 
#define ITC_AddrMask 0 		// 1K address space
#define ITC_EntryGrain 0        // 128 byts between Entries (Cells)

enum vpe {VPE0, VPE1};
enum tc  {TC0, TC1, TC2, TC3, TC4, TC5, TC6, TC7, TC8};

// Declare a reference to the global pointer
extern unsigned int _gp;

// stack areas for TC1 and TC2
unsigned int TC1_stack[4096] __attribute__((aligned(16)));
unsigned int TC1_stack_top = (unsigned int)TC1_stack + 4080;

unsigned int TC2_stack[4096] __attribute__((aligned(16)));
unsigned int TC2_stack_top = (unsigned int)TC2_stack + 4080;

unsigned int TC3_stack[4096] __attribute__((aligned(16)));
unsigned int TC3_stack_top = (unsigned int)TC3_stack + 4080;


// counter 
int TC_Counter[9];

/*
 *------------------------------------------------------------------------
 * This is a simple count loop to increment a cached counter.
 * It will be called by TC0 and TC1
 * After the count reaches 2000 the thread will be yielded
 *------------------------------------------------------------------------
 */
void count() {

	int y, ITC;
	unsigned int *ITC_Cell;
        
    // Try to load from a P/V view of the ITC Cell
    // This will block if the cell contains 0
    ITC_Cell =(unsigned int *)((unsigned int)ITC_BlockNC |ITC_PVSyncView); 
	ITC = *ITC_Cell;

	// Get the TC nuber of this TC
	y= mips32_gettcbind() >> TCBIND_CURTC_SHIFT;
	while(TC_Counter[y] < 2000)
       		TC_Counter[y]=TC_Counter[y] +1;

	// Increment the ITC cell to allow another to enter this code
	*ITC_Cell = ITC;

        // Free up this thread
	mips_mt_yield(0);
}

/*
 *------------------------------------------------------------------------
 * main():  Simple program showing what it takes to kick off vpe0/tc1,tc2. 
 *------------------------------------------------------------------------
 */

int main (int argc, char **argv) {

    // Variables for ITC
    unsigned int errctlreg;
    unsigned int errctlreg_withITC;
    unsigned int ITC_Config_Tag0;
    unsigned int ITC_Config_Tag8;
    unsigned int ITC_CellTag;
    unsigned int *ITC_Cell;

    // Setup ITC

    // Configure the ITC tags. This is done by using cache opts. To set the 
    // cache mode for ITC tags the ITC bit in the ErrCtl register must be set

    errctlreg = mips32_geterrctl();
    errctlreg_withITC = errctlreg | ERRCTL_ITC;
    mips32_seterrctl(errctlreg_withITC);

#if READ_ITC_ENTRIES
    __asm__ volatile
    	("\
    	cache 5, 8($0);  \
    	ehb; \
    	"
     	) ;
#endif

    // configure Number of entries Address mask bits and Entry Grain in ITC tag index 8
    ITC_Config_Tag8 = ((ITC_AddrMask << 10) | ITC_EntryGrain); 
    mips32_setdtaglo(ITC_Config_Tag8);

    __asm__ volatile
    	("\
    	cache 9, 8($0);  \
    	ehb; \
    	"
     	) ;

    // configure Base address and ITC_En (enable bit) in ITC tag index 0 and 

    // Use physical address
    
    ITC_BlockNC = (unsigned int *)((unsigned int)ITC_Block & 0x7fffffff);

    ITC_Config_Tag0 = ((unsigned int)ITC_BlockNC | ITC_En);
    mips32_setdtaglo(ITC_Config_Tag0);

    __asm__ volatile
    	("\
    	cache 9, 0($0); \
    	ehb; \
    	"
     	) ;

    mips32_seterrctl(errctlreg);  // return ErrCtl to it previous state

    // Enable ITC Entry 
    // Use unmapped address
    ITC_FirstPVCell = ITC_Block + (32 *4);
    ITC_BlockNC = (unsigned int *)((unsigned int)ITC_FirstPVCell | 0x20000000);

    ITC_Cell =(unsigned int *)((unsigned int)ITC_BlockNC | ITC_ControlView); 
    ITC_CellTag = *ITC_Cell; 
    *ITC_Cell = ITC_E;

    // Increment the ITC cell to allow the first thread to execute count()
    ITC_Cell =(unsigned int *)((unsigned int)ITC_BlockNC | ITC_PVSyncView); 
    *ITC_Cell = 1;

    // Done setting up ITC

    // Turn on the vpe configuration flag and disable (other) vpe.
    mips32_setmvpcontrol((mips32_getmvpcontrol() & ~MVPCONTROL_EVP) | MVPCONTROL_VPC); // VPC=1 EVP=0.

    // set DA bit so yields will work.
    mips32_mt_settcstatus(mips32_mt_gettcstatus() | TCSTATUS_DA);

    // Set Group Priority to 0
    mips32_settcschedule(0);

/******************************************************************************/
    mips32_mt_settarget(TC1); // Have all 'mips32_mt_*' functions target tc1.

    // Make sure tc1 is halted. (This done in h/w as part of reset.)
    mips32_mt_settchalt(TCHALT_H); // H=1.

    // Make sure tc1 is bound to vpe0. (This is done in h/w as part of reset.)
    mips32_mt_settcbind ((VPE0 << TCBIND_CURVPE_SHIFT) | (TC1 << TCBIND_CURTC_SHIFT));

    // Set the stack pointer so we can call a c function
    mips32_mt_setsp(TC1_stack_top);
    mips32_mt_setgp(&_gp);

    // Point tc1 to code
    mips32_mt_settcrestart(count);

    // Set DA bit so forks/yields will work.
    // Activate tc1 (A=1).
    mips32_mt_settcstatus(mips32_mt_gettcstatus() | (TCSTATUS_A | TCSTATUS_DA));

    // Set Group Priority to 1
    mips32_mt_settcschedule(1);

    // Unhalt tc1.
    mips32_mt_settchalt(0); // H=0 (allow vpe0/tc1 to execute). 

/******************************************************************************/

    mips32_mt_settarget(TC2); // Have all 'mips32_mt_*' functions target tc2.

    // Make sure tc2 is halted. (This done in h/w as part of reset.)
    mips32_mt_settchalt(TCHALT_H); // H=1.

    // Make sure tc2 is bound to vpe0.
    mips32_mt_settcbind ((VPE0 << TCBIND_CURVPE_SHIFT) | (TC2 << TCBIND_CURTC_SHIFT));

    // Set the stack pointer and global pointer so we can call a c function
    mips32_mt_setsp(TC2_stack_top);
    mips32_mt_setgp(&_gp);

    // Point tc2 to code
    mips32_mt_settcrestart(count);

    // Set DA bit so forks/yields will work.
    // Activate tc2 (A=1).
    mips32_mt_settcstatus(mips32_mt_gettcstatus() | (TCSTATUS_A | TCSTATUS_DA));

    // Set Group Priority to 1
    mips32_mt_settcschedule(1);

    // Unhalt tc2.
    mips32_mt_settchalt(0); // H=0 (allow vpe0/tc2 to execute). 

/******************************************************************************/

    mips32_mt_settarget(TC3); // Have all 'mips32_mt_*' functions target tc3.

    // Make sure tc3 is halted. (This done in h/w as part of reset.)
    mips32_mt_settchalt(TCHALT_H); // H=1.

    // Make sure tc3 is bound to vpe0.
    mips32_mt_settcbind ((VPE0 << TCBIND_CURVPE_SHIFT) | (TC3 << TCBIND_CURTC_SHIFT));

    // Set the stack pointer and global pointer so we can call a c function
    mips32_mt_setsp(TC3_stack_top);
    mips32_mt_setgp(&_gp);

    // Point tc2 to code
    mips32_mt_settcrestart(count);

    // Set DA bit so forks/yields will work.
    // Activate tc3 (A=1).
    mips32_mt_settcstatus(mips32_mt_gettcstatus() | (TCSTATUS_A | TCSTATUS_DA));

    // Set Group Priority to 1
    mips32_mt_settcschedule(1);

    // Unhalt tc3.
    mips32_mt_settchalt(0); // H=0 (allow vpe0/tc3 to execute). 



/******************************************************************************/

    // Enable Group Rotation
    mips32_setvpeschedule(0);

    // Enable threading.
    mips32_setvpecontrol(mips32_getvpecontrol() | VPECONTROL_TE); // TE=1.

    // Turn off the vpe configuration flag and enable (other) vpe.
    mips32_setmvpcontrol((mips32_getmvpcontrol() & ~MVPCONTROL_VPC) | MVPCONTROL_EVP); // VPC=0 EVP=1.

    count(0); // Send this tc (vpe0/tc0) into an infinite loop showing its activity.

    return (0); // Never gets here.
}
